/* $Header:   F:/source/cwright/SQL/SQL.c_v   1.42   10 Mar 1993 14:11:52   ericj  $ */
/*
 *
 *                  Copyright 1991,1992 Premia Corp.
 *                        All rights reserved
 *
 *    This source code is provided to the purchaser of the product
 *    for the purpose of allowing the purchaser to customize and/or
 *    extend the functionality of the product.  The source code
 *    remains the property of Premia Corp. with all rights reserved.
 *    Any modifications to the source code are considered derivative
 *    works and all rights thereto are reserved to Premia Corp. except
 *    that the purchaser may use the modified product in the same
 *    permitted manner as the unmodified product.  It is expressly
 *    prohibited to distribute in any manner any part of the original
 *    source code, whether in source or object form, without the
 *    express written permission from Premia Corp.
 *
 *    Owner:
 *        Premia Corporation
 *        1075 NW Murray Blvd., Suite 268
 *        Portland OR 97229
 *
 * History:
 *    1998.02   Chris Bartell   Added User1 and User2 Keyword coloring.
 *                              Added spell checking functionality.
 *    1993.12   Scott Gartner   Modified the file to be SQL.c instead of C.c.  The dialect
 *                              of SQL that this file uses is ClearSQL (copyright ClearAccess
 *                              Corporation).  This should be just fine with any other versions
 *                              of SQL since ClearSQL is a superset of most SQL implementations.
 *                              There is no flow control however, so if you are using an imbeded
 *                              SQL you may have to add some keywords to the hash table.  I
 *                              included some interesting template expansions (although not
 *                              necesarily well thought out).  If you expand the keyword list
 *                              (deleting nothing) then please send me a copy. Compuserve:73352,3361
 *    1993.05   Erik Oxaal      Version 2.0i now incorporates most of chroma-
 *                              additions. Replaced calls to ColorXXX() with
 *                              ColorAlternateX(). My version still handles
 *                              quoted strings better.
 *    1992.10   Erik Oxaal      Extended chroma-coding scheme to include
 *                              C preprocessor directives (ColorPreproc()),
 *                              string and character constants (ColorQuotes()),
 *                              numeric constants (ColorNumbers()),
 *                              and braces,brackets paren. (ColorBraces()).
 *                              (The brace coloring scheme is based on code by
 *                              Terry W. Wilk found on BIX).
 *    1992.10   Erik Oxaal      Changed smart indenting to use CIndentNewl()
 *                              in INDENT.C, a 'better' smart indent! The change
 *                              mostly consists of commenting out unneeded
 *                              code. Codewright's standard method can be
 *                              used by compiling with '-DCWR_INDENT'.
 *    1992.10   Erik Oxaal      Changed '...routines()' so it also includes
 *                              function definitions that have a trailing
 *                              single-line comment.
 *
 *                              THESE CHANGES ARE RELEASED TO BE USED AND/OR
 *                              MODIFIED FREELY BY PREMIA CORP AND REGISTERED
 *                              CODEWRIGHT USERS.
 */

#include <windows.h>
#include <string.h>
#include <ctype.h>
#include <exports.h>
#include "cwstart.h"

#define CWR_INDENT  // Cheat and use the default indenting routines, I'm just lazy

#define MAX_BUF_LEN    100

#define SC_ARG_CNT         6
LPSTR scParms[SC_ARG_CNT] = { "sql", "\"", "/*", "*/", "//", NULL };

static LPVOID _SQL_hash_names        = NIL;  /* pointer to the hash table containing */
                                             /* all 'SQL' keyword names. */

static LPVOID _SQL_user_names_1      = NIL;  /* pointer to the hash table user keyword names. */
static LPVOID _SQL_user_names_2      = NIL;  /* pointer to the hash table user keyword names. */

static LPVOID _SQL_hash_templates    = NIL;  /* pointer to the hash table containing */
                                             /* an abbreviated name and its template */

/* Define a structure to hold an abbreviation and the template to be
 * expanded when the abbreviation is entered into the buffer.
 * This structure will be the element in the hash table _SQL_hash_templates.
 */
typedef struct _LanguageTemplate {
    LPSTR abbrev;
    LPSTR template;
} LanguageTemplate;


/* This string contains all keyword names which will be indented an extra level.
 */
LPSTR SQL_INDENT_STR = "\377if\377while\377else\377do\377for\377select"
                       "\377while\377from\377sort\377group\377create\377drop\377";

/*
 * _SQL_templates[] contains the keyword abbreviation and the template for the
 * abbreviation.  To add a new name, insert it before the NIL entry or
 * call _SQL_assign_template().
 *
 * The special characters within the template are:
 *
 *     \n - new line
 *     &  - cursor position after template insertion
 *     @  - backspace
 *     \c - insert 'c' literally
 */
LanguageTemplate _SQL_templates[] = {
    { "if",                         "if (&)\n{\n}" },
    { "wh|whi|whil|while",          "while (&)\n{\n}" },
    { "do",                         "do\n{\n\t&\n@}\nwhile( )" },
    { "fo|for" ,                    "for (&)\n{\n}" },
    { "el|els|else",                "else\n{\n\t&\n@}" },
    { "ei",                         "else if (&)\n{\n}" },
    { "se|sel|sele|selec|select",   "select &\nfrom " },
    { "cr|cre|crea|creat|create",   "create " },
    { "ta|tab|tabl|table",          "table & ()" },
    { "in|ind|inde|index",          "index & ()" },
    { "gr|gro|grou|group",          "group by &" },
    { NIL,       NIL },
};

/* Define the list of 'SQL' keyword names to appear in
 * the keyword color if language coloring is enabled.
 */
LPSTR _SQL_names[] = {
    "ABS",          "ABSOLUTE",       "ACOS",             "ADD",              "ADDITION",
    "ALIAS",        "ALL",            "ALTER",            "AND",              "ANY",
    "AS",           "ASC",            "ASCII",            "ASIN",             "ATAN",
    "ATAN",         "AVG",            "BETWEEN",          "BOOLEAN",          "BY",
    "CALL",         "CEILING",        "CHAR",             "CHECK",            "CLOSE",
    "COL",          "COLUMN",         "COLUMNS",          "COMMA",            "COMMIT",
    "COMPOP",       "CONCAT",         "CONST",            "CONVERT",          "COS",
    "COT",          "COUNT",          "CREATE",           "CURDATE",          "CURRENT",
    "CURTIME",      "DAE",            "DATABASE",         "DATABASES",        "DATE",
    "DAY",          "DAYOFMONTH",     "DAYOFWEEK",        "DAYOFYEAR",        "DBMS",
    "DBMSBRAND",    "DECIMAL",        "DEFAULT",          "DELETE",           "DESC",
    "DESCRIBE",     "DESELECT",       "DISTINCT",         "DIV",              "DIVISION",
    "DROP",         "EACH",           "EXECUTE",          "EXISTS",           "EXP",
    "EXTENSION",    "EXTRACT",        "FETCH",            "FIRST",            "FLOAT",
    "FLOOR",        "FOR",            "FOREIGN",          "FROM",             "GRANT",
    "GREATEST",     "GROUP",          "HAVING",           "HEADERS",          "HEADERSOF",
    "HOSTS",        "HOUR",           "IN",               "INDEX",            "INITCAP",
    "INSERT",       "INTEGER",        "INTO",             "IS",               "ITEMS",
    "KEY",          "LAST",           "LCASE",            "LEAST",            "LEFT",
    "LENGTH",       "LESS",           "LIKE",             "LIMITED",          "LIST",
    "LOCATE",       "LOCATION",       "LOG",              "LPAD",             "LTRIM",
    "MAX",          "MIN",            "MINUTE",           "MOD",              "MONTH",
    "MONTHS",       "MUL",            "MULTIPLICATION",   "NEG",              "NEW",
    "NEXT",         "NODE",           "NOT",              "NOW",              "NULL",
    "NULLTOKEN",    "NULVALUE",       "NUMBER",           "OF",               "ON",
    "OPEN",         "OPTION",         "OR",               "ORDER",            "PASSWORD",
    "PI",           "POS",            "POWER",            "PREVIOUS",         "PRIMARY",
    "PRINT",        "PRINTALL",       "PRINTCTL",         "PRINTINFO",        "PRINTINFOITEMS",
    "PRINTMAX",     "PRINTMAXOF",     "PRINTNEXT",        "PRINTNEXTOF",      "PRINTROW",
    "PRIVILEGES",   "PUBLIC",         "QUARTER",          "RAND",             "READONLY",
    "REAL",         "REFERENCES",     "RELATIVE",         "REPEAT",           "REPLACE",
    "RESOLVE",      "REVOKE",         "RIGHT",            "ROLLBACK",         "ROUND",
    "RPAD",         "RTRIM",          "SCROLLING",        "SECOND",           "SELECT",
    "SET",          "SIGN",           "SIN",              "SMFLOAT",          "SMINT",
    "SOME",         "SOUNDEX",        "SQL",              "SQRT",             "STDDEV",
    "STRING",       "SUB",            "SUBSTRING",        "SUBTRACTION",      "SUM",
    "TABLE",        "TABLES",         "TAN",              "TIME",             "TIMESTAMP",
    "TO",           "TRUNC",          "UCASE",            "UNIQUE",           "UPDATE",
    "USE",          "USER",           "VALUES",           "VARBIN",           "VARCHAR",
    "VARIANCE",     "VIEW",           "WEEK",             "WHERE",            "WITH",
    "WORK",         "YEAR",
    NIL,
};


static long _SQL_in_comment( long offset );
static long _SQL_find_comment_start( void );
static long _SQL_find_comment_end( UINT style );

void DLL _SQL_init( void );
BOOL DLL _SQL_expand( void );
long DLL _SQL_language_colors_update( long firstline, long lastline );
void DLL _SQL_language_colors( long numlines );
void DLL _SQL_assign_template( LPSTR abbrev, LPSTR template );
void DLL _SQL_add_keyword( LPSTR keyword, int table );
BOOL DLL _SQL_indent( void );
void DLL _sql_spellcheck_data_request(int mode);


#ifndef CWR_INDENT
extern void DLL CIndentNewl( void );
#endif

LIBMAIN

/*
 ** _SQL_init()
 *
 *  Initialize this module by exported any required functions and
 *  the hash tables needed to support indenting, template expandion
 *  and keyword coloring.
 */
void DLL
_init( void )
{
    static BOOL firstTime = TRUE;
    int    i;

    if (firstTime)
    {
      LibExport( "BOOL _SQL_expand");
      LibExport( "_SQL_assign_template LPSTR LPSTR");
      LibExport( "_SQL_add_keyword LPSTR int");
      LibExport( "BOOL _SQL_indent( void ) ");
      LibExport( "long _SQL_language_colors_update long long");
      LibExport( "_SQL_language_colors long");
      LibExport( "_SQL_routines" );
	  LibExport( "_sql_spellcheck_data_request int ");

#ifndef CWR_INDENT
      LibExport( "CIndentInit int int" );
#endif

      firstTime = FALSE;

      /* Create a hash table to hold all of the 'SQL' keyword names.
       */
      _SQL_hash_names = HashCreateTable( 41 /* a prime # */, sizeof( LPSTR ), 0 );
      i = 0;

      while ( _SQL_names[ i ] != NIL )
      {
        LPSTR XFAR *lpf = (LPSTR XFAR *)HashGetEntry( _SQL_hash_names,
                  _SQL_names[i], strlen( _SQL_names[i] ) );
        if (lpf)
        {
          *lpf = StrNew( _SQL_names[i] );
          i++;
        }
      }

      /* Create a hash table to hold all of the 'SQL' template names.
       */
      _SQL_hash_templates = HashCreateTable( 17 /* a prime # */, sizeof( LPSTR ), 0 );
      i = 0;

      while ( _SQL_templates[ i ].abbrev != NIL )
      {
        _SQL_assign_template( _SQL_templates[i].abbrev, _SQL_templates[i].template );
        i++;
      }
	  CompilerNewExt( ".sql" );

    }
}


/*
 ** _SQL_expand()
 *
 *  DESCRIPTION:
 *    Attempts to perform 'SQL' language template expansion.  This function
 *    is normally bound to the <SPACE> key.  When invoked, it reads the
 *    previous word.  If the word is in the list of defined templates,
 *    the template is extracted and inserted.  Some characters in the
 *    templates have special meanings. They are:
 *
 *       \n - new line
 *       &  - cursor position after template insertion
 *       @  - backspace
 *       \  - take next character literally
 */
BOOL DLL
_SQL_expand( void )
{
    LPVOID htEntry;
    LPSTR  _SQL_string;
    LPSTR  str;

    MarkSavePos();

    /* Check to see if we are currently positioned inside a 'SQL' comment
     */
    if (_SQL_in_comment( -1L ) != -1)
      goto _SQL_expand_incomplete;

    str = BufReadStr( 65535 );     /* read entire line */

    _SQL_string = StrTrim( str, " \t" );
    StrFree( str );

    if (!_SQL_string)
      goto _SQL_expand_incomplete;

    /* Search for a keyword match in the hash table.
     */
    htEntry = HashFindEntry( _SQL_hash_templates, _SQL_string, strlen( _SQL_string ) );
    StrFree( _SQL_string );

    if (htEntry)
    {
      /* A template exists for the current keyword, delete the keyword
       * and insert the template.
       */
      MarkDropPos( 1 );
      str = *(LPSTR*)htEntry;
      BufDelToEOL();

      _ext_expand_template( str );

    }
    else
      goto _SQL_expand_incomplete;
    return TRUE;

_SQL_expand_incomplete:
    MarkRestorePos();

    return FALSE;
}


/*
 ** _SQL_language_colors_update()
 *
 *  This function does an intermediate update of the colors within
 *  a range of line numbers (firstline and lastline).  It must also
 *  determine if the first and last line specified are within a
 *  comment and act appropriately.
 */
long DLL
_SQL_language_colors_update( long firstline, long lastline )
{
    long startoff;
	long count = 0;

    long eSQL_line  = 0; /* end comment line */
    UINT style    = 0;

    if (firstline && lastline)
    {
      /* Begin coloring at the previous comment if
       * currently located in the middle of one.
       */
      _PosInit( 0 );
      if (firstline == 1)
        _PosPrevLine( 0L );
      else
        _PosNextLine( firstline - 1);

      startoff = _PosQOffset();
      if ( _SQL_find_comment_start() == 1)
      {
        startoff  = _PosQOffset();

        _PosNextChar();   /* move off of the comment */
        _PosNextChar();

        style = 1;

        if (_SQL_find_comment_end( style ))
          eSQL_line = _PosQLine();

        _PosInit( startoff );
        firstline = _PosQLine();
      }

      /* End coloring at the next comment if
       * currently located in the middle of one.
       */
      _PosInit( 0 );
      _PosNextLine( lastline - 1 );
      _PosNextLine( 0 );

      if (_PosQColor())
      {
        _SQL_find_comment_end( style );
        lastline = _PosQLine();
      }
      else if (_SQL_find_comment_end( style ) == 2)
        lastline = _PosQLine();


      if (!eSQL_line)
	  {
        AttrSetColor( firstline, lastline, 1L, 0x7FFFFFFFL, 0 );
	 	count = lastline - firstline + 1;

	  }
      else if (eSQL_line < lastline)
	  {
        AttrSetColor( eSQL_line, lastline, 1L, 0x7FFFFFFFL, 0 );
	  	count = lastline - eSQL_line + 1;
	  }


      /* Reset the attributes for the line range within _SQL_language_colors.
       */
      _PosInit( startoff );

      _SQL_language_colors( count );
    }
	return count;
}


/*
 ** _SQL_language_colors()
 *
 * If numlines == -1, start from the top of the buffer and encode all of them,
 * otherwise start at the current location of the number of lines specified.
 *
 * It is assumed that if numlines != -1, that _PosInit() has been called
 * and numlines is relative to the current position.
 */
void DLL
_SQL_language_colors( long numlines )
{
    long  startoff;
    long  startline;
    long  endoff;
    long  inc;
    UINT  ch;
    LPSTR buf;
    int   i = 0;
    int   comment_color    = ColorComments( -1 );
    int   keyword_color    = ColorKeywords( -1 );
    long  maxlines         = ExtCommentSearchLimit( -1 );
    int   quote_color      = ColorAlternate1( -1 );
    int   number_color     = ColorAlternate2( -1 );
    int   brace_color      = ColorAlternate3( -1 );
    int   preproSQL_color    = ColorAlternate4( -1 );
	int   color;

    buf = (LPSTR)MemAlloc( MAX_BUF_LEN );

    /* Start at the top of the buffer or current position
     */
    if (numlines == -1)
    {
      _PosInit( 0 );
      numlines = 0x7FFFFFFF;
    }
    else
    {
      startoff  = _PosQOffset();
      startline = _PosQLine();
      endoff    = _SQL_in_comment( startoff );
      if (endoff == -1)
        _PosInit( startoff );
      else
      {
        _PosInit( endoff );
        numlines += (startline - _PosQLine());
      }
    }
    ch = _PosCurrentChar();
    startoff = _PosQOffset();

    maxlines = max( numlines, maxlines );

    while ((ch != EOF_CHAR) && (numlines > 0))
    {
      if (ch == '\n')
        numlines--;

      if (ch == '"' || ch == '\'')
      {
        UINT quote = ch;

            startoff = _PosQOffset();

        do
        {
          ch = _PosNextChar();
          if (ch == '\\')
          {
            ch = _PosNextCharNL();
            if ((ch == quote) || (ch == '\\') || (ch == EOL1_CHAR) || (ch == EOL2_CHAR))
              ch = _PosNextCharNL();
          }
          if ((ch == EOF_CHAR) || (ch == '\n'))
            break;
        } while (ch != quote);

        if (ch == quote)
          ch = _PosNextChar();

        /* color the string */
        if (quote_color)
        {
          endoff = _PosQOffset();
            _PosInit( startoff );
          _PosSetColor( quote_color, endoff - startoff );
          _PosInit( endoff );
        }
        continue;
      }
      else if (number_color && isdigit( ch ))
      {
        int j = 0;
        UINT lastch = _PosPrevChar();

        if (lastch != EOF_CHAR)
          _PosNextChar();

        startoff = _PosQOffset();
        do
        {
          j++;
          ch = _PosNextChar();
        } while (isxdigit( ch ) || (ch == '.') || (ch == 'x' && j == 1));

        if ((  lastch != ' '
          && lastch != '\t'
          && lastch != '\r'
          && lastch != '\n'
          && lastch != EOF_CHAR
          && !ispunct( lastch ))
          || lastch == '_'
          || (ch == '_'))
          continue;

        if (ch == 'l' || ch == 'L')
           ch = _PosNextChar();

        endoff = _PosQOffset();
        _PosInit( startoff );
        _PosSetColor( number_color, endoff - startoff );
        _PosInit( endoff );

      }
      else if ( ch == '#' && preproSQL_color)
      {
        startoff = _PosQOffset();
        do
        {
          ch = _PosNextChar();
        } while ( ch == ' ' || ch == '\t');
        do
        {
          ch = _PosNextChar();
        } while ( !isspace(ch) && ch != EOF_CHAR);
        endoff = _PosQOffset();
        _PosInit( startoff );
        _PosSetColor( preproSQL_color, endoff - startoff );
        _PosInit( endoff );
      }
      else if (isalpha( (BYTE)ch ) || (ch == '_'))
      {
        UINT lastch = _PosPrevChar();

        if (lastch != EOF_CHAR)
          _PosNextChar();

        i = 0;
        startoff = _PosQOffset();

        do
        {
          if (i < MAX_BUF_LEN)
            buf[i++] = (char)ch;
          ch = _PosNextChar();
        } while (ISALNUM( (BYTE)ch ) || (ch == '_'));

        if ((lastch != ' '
              && lastch != '\t'
              && lastch != '\r'
              && lastch != '\n'
              && lastch != EOF_CHAR
              && !ispunct( lastch ))
            || (ch == '_'))
          i = 0;
        continue;
      }
      else if (i)
      {
 			STRUPR(buf);
			if (_SQL_user_names_1 && HashFindEntry( _SQL_user_names_1, buf, i ))
				color = ColorAlternate5( -1 );
			else if (_SQL_user_names_2 && HashFindEntry( _SQL_user_names_2, buf, i ))
				color = ColorAlternate6( -1 );
			else if (HashFindEntry( _SQL_hash_names, buf, i ))
				color = keyword_color;
			else
				color = 0;

			if (color)
			{
				endoff = _PosQOffset();
				_PosInit( startoff );
				_PosSetColor( color, endoff - startoff );
				_PosInit( endoff );
			}
       		i = 0;
      }

      if (ch == '/')
      {
        startoff = _PosQOffset();
        ch = _PosNextChar();

        if (ch == '/')
        {
          inc = _PosNextLine( 1 );
          numlines--;
          endoff = _PosQOffset();
          _PosInit( startoff );
          _PosSetColor( comment_color, endoff - startoff - inc );
          _PosInit( endoff );
          ch = _PosCurrentChar();
        }
        else if (ch == '*')
        {
          long lines = 0;

          ch = _PosNextChar();
          while ((ch != EOF_CHAR) && (lines < maxlines))
          {
            if (ch == '\n')
            {
              lines++;
              numlines--;
            }
            else if (ch == '*')
            {
              ch = _PosNextChar();
              if (ch == '/')
              {
                endoff = _PosQOffset();
                _PosInit( startoff );
                _PosSetColor( comment_color, endoff - startoff + 1 );
                _PosInit( endoff );
                break;
              }
              else
                continue;
            }
            ch = _PosNextChar();
          }
          ch = _PosNextChar();
        }
        continue;
      }
      else if (brace_color && _fstrchr("{}[]()", ch) != 0)
      {
        long coff;

        coff = _PosQOffset();
        _PosInit(coff);
        _PosSetColor(brace_color, 1);
        _PosInit(coff);
      }
      ch = _PosNextChar();

    }
    MemFree( buf );
}



void DLL
_SQL_add_keyword( LPSTR keyword, int table  )
{
	if (keyword && *keyword)
	{
		LPSTR XFAR *lpf;
		LPVOID hash_table;

		switch( table )
		{
		case 0:
			hash_table = _SQL_hash_names;
			break;
		case 1:
			if (!_SQL_user_names_1)
				_SQL_user_names_1 = HashCreateTable( 41, /* a prime # */
						sizeof( LPSTR ), 0 );
			hash_table = _SQL_user_names_1;
			break;
		case 2:
			if (!_SQL_user_names_2)
				_SQL_user_names_2 = HashCreateTable( 41, /* a prime # */
						sizeof( LPSTR ), 0 );
			hash_table = _SQL_user_names_2;
			break;
		default:
			hash_table = NIL;
			break;
		}
		if (hash_table)
		{
			lpf = (LPSTR XFAR *)HashGetEntry( hash_table, 
					STRUPR(keyword), STRLEN( keyword ) );
			if (lpf)
				*lpf = StrNew( keyword );
		}
	}
}


/*
 ** _SQL_assign_template()
 *
 *  PARAMETERS:
 *    abbrev   - the keyword abreviation used to locate a template
 *    template - the encoded template string associated with 'abbrev'
 *
 *  DESCRIPTION:
 *    Provide a means of adding or modifying 'SQL' template strings
 *    without changing the code.
 */
void DLL
_SQL_assign_template( LPSTR abbrev, LPSTR template )
{
    _ext_assign_template( _SQL_hash_templates, abbrev, template );
}


/*
 ** _SQL_indent()
 *
 *  DESCRIPTION:
 *    Attempts to perform 'SQL' language smart indenting.  This function
 *    will behave differently based on the following conditions:
 *
 *    1. Within a comment - normal indentation
 *
 *    2. Not positioned at end of line - normal indentation
 *
 *    3. Current line prefixed by a name in SQL_INDENT_STR - perform smart
 *       indentation.  This may include skipping the next line if
 *       it consists only of an open brace.
 */
BOOL DLL
_SQL_indent( void )
{
    long   curline;
    long   col;
    LPSTR  _SQL_string;
    LPSTR  str;
    long   matchpos;
    UINT   endch;

    MarkSavePos();

    curline = BufQCurrentLine();

    /* Check to see if we are currently positioned inside a 'SQL' comment
     */
    if (_SQL_in_comment( -1L ) != -1)
      goto _SQL_indent_incomplete;

    /* Check to see if there is a trailing '{' on the current line,
     * and if so, indent on the next line.
     */
    do {
      MovPrevChar( 1L );
      endch = BufReadChar();
    } while (endch == ' ' || endch == '\t');
    MarkRestorePos();
    if (endch == '{')
      goto _SQL_indent_next_line;
    MarkSavePos();
#ifdef CWR_INDENT

    /* Not in a comment so indent according to the current 'SQL' construct.
     */
    if (!SrchFind( "^[ \t]*[a-zA-Z_]+\\c([ \t]|$)", SEARCH_REGEX , NIL )
        || (BufQCurrentLine() != curline) )
      goto _SQL_indent_incomplete;

    /* Read the keyword at the beginning of the current line.
     */
    col = BufQCurrentCol();
    SrchFind( "(^|[^a-zA-Z_])\\c[a-zA-Z_]", SEARCH_REGEX , NIL );
    str = BufReadStr( col - BufQCurrentCol() );
    _SQL_string = StrTrim( str, " \t" );
    StrFree( str );

    /* Set prefix and suffix to \377 to perform a quick lookup of the name
     */
    str = StrNew( "\377" );
    _SQL_string = StrAppend( _SQL_string, str );
    str = StrAppend( str, _SQL_string );
    StrFree( _SQL_string );

    matchpos = StrMatch( str, SQL_INDENT_STR, SEARCH_FORWARD, NIL );
    StrFree( str );

    if (!matchpos)
      /* Not a 'SQL' keyword, do normal indentation. */
      goto _SQL_indent_incomplete;

    MarkRestorePos();
    MarkSavePos();
#endif
    /* Check to see if a '{' is by itself on the next line. If so, skip
     * it an indent on the line following it.
     */
    if (SrchFind( "^[ \t]*{\\c[ \t]*$", SEARCH_FORWARD | SEARCH_REGEX, NIL )
        && (BufQCurrentLine() == curline + 1) )
      MarkDropPos( 1 );
    else
      MarkRestorePos();

_SQL_indent_next_line:
#ifdef CWR_INDENT
    _ext_insert_indented_EOL();
    BufInsertChar( '\t' );
#else
    CIndentNewl();  //eox
#endif
    return TRUE;

_SQL_indent_incomplete:
    MarkRestorePos();

    return FALSE;
}



/*--  Private functions --*/


/*
 ** _SQL_in_comment()
 *
 *  Determines whether or not the current position in the buffer is
 *  located within a 'SQL' comment by search backward for a // or /*
 *  comment declaration.
 *
 *  Returns -1 if not in a comment
 *          >= 0 indicates the offset in the buffer where the comment begins
 */
static long
_SQL_in_comment( long offset )
{
    UINT  ch;
    BOOL  firstline = TRUE;
    long  lines     = 0;
    long  maxlines = ExtCommentSearchLimit( -1 );

    /* Check to see if we are currently positioned inside a 'SQL' comment
     */
    _PosInit( offset );
    ch = _PosCurrentChar();

    do
    {
      switch( ch )
      {
      case '*':
        ch = _PosPrevChar();
        if (ch == '/')    /* Found the beginning of a comment */
        {
          offset = _PosQOffset();  /* currrent line is in a comment */
          ch = _PosPrevChar();
          if ((ch != '/') || (ch == '/' && firstline))
          {
            _PosInit( offset );
            return offset;
          }
        }
        break;
      case '/':
        ch = _PosPrevChar();
        if (ch == '/')
        {
          if (firstline)
            return _PosQOffset();  /* currrent line is in a comment */
          else
            return -1;    /* Not in a comment */
        }
        else if (ch == '*')
          return -1L;    /* Not in a comment */
        break;

      case '"':
        ch = _PosPrevChar();
        while (ch != EOF_CHAR && ch != '"' && ch != '\n')
          ch = _PosPrevChar();
        ch = _PosPrevChar();
        break;

      case '\n':
        lines++;
        firstline = FALSE;
      default:
        ch = _PosPrevChar();
        break;
      }
    } while ((ch != EOF_CHAR) && (lines < maxlines));

    return -1L;
}

/*
 ** _SQL_find_comment_start()
 *
 *  Search backwards for a 'SQL' comment sequence. Return 1 if /<*> found
 *  and 2 if <*>/ found, 0 otherwise. Leave the current position
 *  at the beginning of the found comment.
 */
static long
_SQL_find_comment_start( void )
{
    UINT ch    = _PosCurrentChar();
    long  lines = 0;
    long  maxlines = ExtCommentSearchLimit( -1 );

    do
    {
      if (ch == '*')
      {
        ch = _PosPrevChar();
        if (ch == '/')
          return 1;
      }
      else if (ch == '/')
      {
        ch = _PosPrevChar();
        if ((ch == '/') && (lines == 0))
          break;
        if (ch == '*')
        {
          long offset = _PosQOffset();  /* currrent line is in a comment */
          ch = _PosPrevChar();
          if (ch == '/')
            return 1;
          _PosInit( offset );
          return 2;
        }
      }
      else
      {
        if (ch == '\n')
          lines++;
        else if (ch == '"')
        {
          ch = _PosPrevChar();
          while (ch != EOF_CHAR && ch != '"' && ch != '\n')
            ch = _PosPrevChar();
        }
        ch = _PosPrevChar();
      }
    } while ((ch != EOF_CHAR) && (lines < maxlines));

    return 0;
}

/*
 ** _SQL_find_comment_start()
 *
 *  Search forwards for a 'SQL' comment sequence. Return 1 if /<*> found
 *  and 2 if <*>/ found, 0 otherwise.  Leave the current position
 *  at the end of the found comment.
 */
static long
_SQL_find_comment_end( UINT style )
{
    long lines = 0;
    UINT ch    = _PosCurrentChar();
    long  maxlines = ExtCommentSearchLimit( -1 );

    do
    {
      if (ch == '*')
      {
        ch = _PosNextChar();
        if (ch == '/')
          return 2;
      }
      else if (ch == '/')
      {
        ch = _PosNextChar();
        if ((ch == '/') && style)
          break;
        if (ch == '*')
          return 1;
      }
      else
      {
        if (ch == '\n')
          lines++;
        else if (ch == '"')
        {
          ch = _PosNextChar();
          while (ch != EOF_CHAR && ch != '"' && ch != '\n')
            ch = _PosNextChar();
        }
        ch = _PosNextChar();
      }
    } while ((ch != EOF_CHAR) && (lines < maxlines));

    return 0;
}





void DLL
_SQL_routines( void )
{
    int   matches = 0;
    LPSTR str     = "^[a-zA-Z0-9_].*\\((.|\n)*\\)";
    LPSTR stre    = "^[a-zA-Z0-9_].*\\((.|\n)*\\)[ \t\n]*\\c[^ \t\n]";
    long  curline;
    UINT  ch;
    long  lines;

#ifdef LIGHT
    if (Grep( "^[a-zA-Z0-9_].*\\)[ \t]*{?([ \t]*((/\\*.*\\*/)|(//.*))?[ \t]*)$" ))
      MsgMessage( "" );
    else
      MsgMessage( "No routines found." );
#else
    SysWaitStart();

    MarkSavePos();

    BufSetAscii();
    MovTopBuf();

    while (SrchFind( str, SEARCH_REGEX | SEARCH_FORWARD, 0 ))
    {
      curline = BufQCurrentLine();

      SrchFind( stre, SEARCH_REGEX | SEARCH_FORWARD, 0 );

      ch = BufReadChar();
      if (ch != ';')
      {
        if (!matches)
          AttrSetVisible( 0, 0, 0 );
        matches++;
        lines = BufQCurrentLine() - curline
              - !!((ch == '{') && (BufQCurrentCol() == 1));
        AttrSetVisible( curline, curline + (lines < 0L ? 0L : lines), 1 );
        MovHome();
      }

      if (!MovDown( 1 ))
        break;
    }

    if (matches)
    {
      BufSetCompact();
      MarkDropPos( 1 );
      MovTopBuf();
    }
    else
      MarkRestorePos();

    SysWaitEnd();

    MsgMessage( matches ? "" : "No routines found." );
#endif
}


void DLL
_sql_spellcheck_data_request(int mode) 
{
	if (mode == 0)
		LibFunctionExecArgv("ExtAddSpellComments", SC_ARG_CNT, scParms);
}
